
/* ###--------------------------------------------------------------### */
/* file		: c_mips_r3000.c					*/
/* date		: Mar 25 1997						*/
/* version	: v3.0							*/
/* author	: Pirouz Bazargan Sabet					*/
/* content	: C model of a circuit					*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include BEH_H
#include "vh_ltype.h"
#include "vh_lspec.h"
#include "vh_xspec.h"

/* ###--------------------------------------------------------------###	*/
/* version	: v0.8							*/
/* origin	: this description has been developed by the ASIM team	*/
/*		  of LIP6 laboratory at University Pierre et Marie	*/
/*		  Curie 4 Place Jussieu 75252 Paris Cedex 05 - France	*/
/* descr.	: functionnal description of a Mips R3000 pipeline	*/
/*		  processor						*/
/* authors	: P. Bazargan Sabet					*/
/* ###--------------------------------------------------------------###	*/

/* ###--------------------------------------------------------------###	*/
/*    local static variables (registers and memory elements)		*/
/* ###--------------------------------------------------------------###	*/

struct local_var
  {
  unsigned int LAST_CK    ;	/* last clock value			*/
  unsigned int NEXTPC     ;	/* next instruction address		*/
  unsigned int REDOPC     ;	/* next instruction address		*/
  unsigned int R     [32] ;	/* integer             registers file	*/
  unsigned int LO         ;	/* low                 register		*/
  unsigned int HI         ;	/* high                register		*/
  unsigned int CAUSE      ;	/* cause               register (cop 0)	*/
  unsigned int NEXTSR     ;	/* status              register (cop 0)	*/
  unsigned int BADVADR    ;	/* bad virtual address register (cop 0)	*/
  unsigned int EPC        ;	/* exception program   counter  (cop 0)	*/
  unsigned int I_FETCH    ;	/* instruction fetch			*/
  unsigned int IFC_BUSY   ;	/* instruction fetch stage busy		*/
  unsigned int INST       ;	/* instruction				*/
  unsigned int OPCOD      ;	/* operation code			*/
  unsigned int SHORT_I    ;	/*          short immediate		*/
  unsigned int BRNCH_DEL  ;	/* branch delayed slot instruction	*/
  unsigned int SHIFT_AM   ;	/* shift amount				*/
  unsigned int RS_NBR     ;	/* first  source     register's number	*/
  unsigned int RT_NBR     ;	/* second source     register's number	*/
  unsigned int RD_NBR     ;	/* destination       register's number	*/
  unsigned int RD_BNBR1   ;	/* destination       register's number	*/
  unsigned int RD_BNBR2   ;	/* destination       register's number	*/
  unsigned int RESULT     ;	/* operation's result			*/
  unsigned int MEM_BUSY   ;	/* memory access     stage busy		*/
  unsigned int MEM_OPCOD  ;	/* operation code's in memory stage	*/
  unsigned int MEM_OP     ;	/* operation type   in memory stage	*/
  unsigned int MEM_RD_NBR ;	/* destination       register's number	*/
  };

void c_mips_r3000 (pt_lkdins)

struct lkdins *pt_lkdins;

  {
	/* ###------------------------------------------------------###	*/
	/*    initialization function					*/
	/* ###------------------------------------------------------###	*/

  static void  init_r3000  ();
  static void  display_var ();

	/* ###------------------------------------------------------###	*/
	/*    circuit's ports :						*/
	/*      - each port is declared as an unsigned integer		*/
	/*      - the relative position of input ports is defined	*/
	/* ###------------------------------------------------------###	*/

  unsigned int ck       ;		/* I         clock		*/
  unsigned int reset_n  ;		/* I         reset		*/
  unsigned int it_n     ;		/* I   5-0   interrupts	*/
/*unsigned int i_a      ;	*/	/* O  31-0   instruction adr	*/
  unsigned int i_ftch   = 1;		/* O         instruction fetch	*/
  unsigned int i_berr_n ;		/* I         inst. bus error	*/
  unsigned int i_frz    ;		/* I         i-mem not ready	*/
/*unsigned int i        ;	*/	/* I  31-0   instruction bus	*/
/*unsigned int d_a      ;	*/	/* O  31-0   data address	*/
  unsigned int d_rq     = 0;		/* O         data access request*/
  unsigned int d_lock   = 0;		/* O         lock memory bus	*/
  unsigned int d_atype  = 0;		/* O   1-0   data access type	*/
  unsigned int d_berr_n ;		/* I         data  bus error	*/
  unsigned int d_frz    ;		/* I         d-mem not readdy	*/
  unsigned int d_in     ;		/* I  31-0   read  data		*/
  unsigned int d_out    ;		/* O  31-0   write data		*/
/*unsigned int test     ;	*/	/* I         test mode		*/
/*unsigned int scin     ;	*/	/* I         scan in		*/
/*unsigned int scout    ;	*/	/* O         scan out		*/
/*unsigned int vdd      ;	*/	/* I				*/
/*unsigned int vss      ;	*/	/* I				*/
/*unsigned int vddp     ;	*/	/* I				*/
/*unsigned int vssp     ;	*/	/* I				*/

#define PI_vssp                     0	/* I				*/
#define PI_vddp      PI_vssp     +  1	/* I				*/
#define PI_vss       PI_vddp     +  1	/* I				*/
#define PI_vdd       PI_vss      +  1	/* I				*/
#define PI_scin      PI_vdd      +  1	/* I				*/
#define PI_test      PI_scin     +  1	/* I				*/
#define PI_d_in      PI_test     +  1	/* I  31-0			*/
#define PI_d_frz     PI_d_in     + 32	/* I				*/
#define PI_d_berr_n  PI_d_frz    +  1	/* I				*/
#define PI_i         PI_d_berr_n +  1	/* I  31-0			*/
#define PI_i_frz     PI_i        + 32	/* I				*/
#define PI_i_berr_n  PI_i_frz    +  1	/* I				*/
#define PI_it_n      PI_i_berr_n +  1	/* I   5-0			*/
#define PI_reset_n   PI_it_n     +  6	/* I				*/
#define PI_ck        PI_reset_n  +  1	/* I				*/

#define PO_scout                    0	/* O				*/
#define PO_d_out     PO_scout    +  1	/* O  31-0			*/
#define PO_d_atype   PO_d_out    + 32	/* O   1-0			*/
#define PO_d_lock    PO_d_atype  +  2	/* O				*/
#define PO_d_rq      PO_d_lock   +  1	/* O				*/
#define PO_d_a       PO_d_rq     +  1	/* O  31-0			*/
#define PO_i_ftch    PO_d_a      + 32	/* O				*/
#define PO_i_a       PO_i_ftch   +  1	/* O  31-0			*/

	/* ###------------------------------------------------------###	*/
	/*    internal variables					*/
	/* ###------------------------------------------------------###	*/

  unsigned int new_inst    ;		/* new instruction		*/
  unsigned int opcod       ;		/* complete operation code	*/
  unsigned int short_i     ;		/*          short immediate	*/
  unsigned int rs_nbr      ;		/* first  source register's nbr	*/
  unsigned int rs_bnbr     ;		/* first  source register's nbr	*/
  unsigned int rt_nbr      ;		/* second source register's nbr	*/
  unsigned int rt_bnbr     ;		/* second source register's nbr	*/
  unsigned int rd_nbr      ;		/* destination   register's nbr	*/
  unsigned int rd_bnbr1    ;		/* destination   register's nbr	*/
  unsigned int rd_bnbr2    ;		/* destination   register's nbr	*/
  unsigned int shift_am    ;		/* shift amount			*/
  unsigned int b_delay     = 0;		/* branch delayed slot		*/
  unsigned int nextpc      ;		/* next instruction address	*/
  int          s_oper      ;		/* first  source      operand	*/
  int          t_oper      ;		/* second source      operand	*/
  unsigned int result      ;		/* execution's result		*/
  unsigned int exc_rq      = 0;		/* exception request		*/

  struct local_var *pt_mem ;		/* local ststic variables	*/

	/* ###------------------------------------------------------###	*/
	/*   instruction set table:					*/
	/*     Opcods in lower case are MIPS R3000 instructions		*/
	/*     Opcods in upper case are application specific		*/
	/*								*/
	/*								*/
	/*   primary opcod (31 downto 26):				*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 |speci|bcond|  j  | jal | beq | bne |blez |bgtz |	*/
	/*    1 |addi |addui|slti |sltui|andi | ori |xori | lui |	*/
	/*    2 |cop0 |  +  |  +  |  +  |     |  +  |     |     |	*/
	/*    3 |  +  |  +  |     |     |     |     |     |     |	*/
	/*    4 | lb  | lh  |  +  | lw  | lbu | lhu |  +  |SWAP |	*/
	/*    5 | sb  | sh  |  +  | sw  |     |     |  +  |     |	*/
	/*    6 |  +  |  +  |  +  |  +  |  +  |  +  |  +  |  +  |	*/
	/*    7 |  +  |  +  |  +  |  +  |  +  |  +  |  +  |  +  |	*/
	/*								*/
	/*								*/
	/* special opcod extension (5 downto 0):			*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 | sll |     | srl | sra |sllv |     |srlv |srav |	*/
	/*    1 | jr  |jalr |     |     |sysca|break|     |SLEEP|	*/
	/*    2 |  +  |  +  |  +  |mtlo |     |     |     |     |	*/
	/*    3 |  +  |  +  |  +  |  +  |     |     |     |     |	*/
	/*    4 | add |addu | sub |subu | and | or  | xor | nor |	*/
	/*    5 |     |     | slt |sltu |     |     |     |     |	*/
	/*    6 |     |     |     |     |  +  |  +  |  +  |  +  |	*/
	/*    7 |     |     |     |     |  +  |  +  |  +  |  +  |	*/
	/*								*/
	/*								*/
	/* bcond opcod extension (20 downto 16):			*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 |bltz |bgez |     |     |     |     |     |     |	*/
	/*    1 |     |     |     |     |     |     |     |     |	*/
	/*    2 |bltza|bgeza|     |     |     |     |     |     |	*/
	/*    3 |     |     |     |     |     |     |     |     |	*/
	/*								*/
	/*								*/
	/* cop0 opcod extension (22, 21, 16 / 25, 24, 23):		*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 | mf  | mt  |  +  |  +  | c0  | c0  | c0  | c0  |	*/
	/*    1 | mf  | mt  |  +  |  +  | c0  | c0  | c0  | c0  |	*/
	/*    2 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*    3 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*    4 |  +  |  +  |     |     | c0  | c0  | c0  | c0  |	*/
	/*    5 |  +  |  +  |     |     | c0  | c0  | c0  | c0  |	*/
	/*    6 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*    7 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*								*/
	/*								*/
	/* c0 cop0 extension extension (4 downto 0):			*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 |  +  |  +  |  +  |     |     |     |  +  |     |	*/
	/*    1 |  +  |     |     |     |     |     |     |     |	*/
	/*    2 | rfe |     |     |     |     |     |     |     |	*/
	/*    3 |  +  |     |     |     |     |     |     |     |	*/
	/*								*/
	/* ###------------------------------------------------------###	*/

#define D_NOACCESS   0x00000000		/* no data access		*/
#define D_ACCESS     0x00000001		/* request for a data access	*/

#define D_READ       0x00000001		/* read request			*/
#define D_WRITEB     0x00000003		/* byte      write request	*/
#define D_WRITEH     0x00000002		/* half word write request	*/
#define D_WRITEW     0x00000000		/* word      write request	*/
#define D_SWAPRQ     0x00000001		/* swap            request	*/

#define I_NOACCESS   0x00000000		/* no instruction fetch		*/
#define I_ACCESS     0x00000001		/* request for instruction fetch*/

#define EXC_RQ       0x00000040		/* exception request		*/
#define SYSCALL_EXC  0x00000060		/* system call     exception	*/
#define BREAK_EXC    0x00000064		/* break point     exception	*/
#define RESERVE_EXC  0x00000068		/* reserved instr. exception	*/
#define CPU_EXC      0x0000006c		/* copro. unusable exception	*/
#define OVR_EXC      0x00000070		/* overflow        exception	*/
#define ADES_EXC     0x00000054		/* store address   exception	*/
#define ADEL_EXC     0x00000050		/* load  address   exception	*/

#define OP_R2R       0			/* register operation		*/
#define OP_LOAD      1			/* load     operation		*/
#define OP_STORE     2			/* store    operation		*/
#define OP_SWAP      3			/* swap     operation		*/
#define OP_BRANCH    4			/* swap     operation		*/

#define NOP_I        0x00000021		/* nop instruction (addu 0,0,0)	*/
#define EXCPHND_A    0x80000080		/* excption   handler address	*/
#define BOOTHND_A    0xbfc00180		/* boot strap handler address	*/
#define RESET_A      0xbfc00000		/* reset vector			*/

#define CAUSE_S      13			/* cause   register's number	*/
#define BADVADR_S     8			/* badvadr register's number	*/
#define EPC_S        14			/* epc     register's number	*/
#define STATUS_S     12			/* status  register's number	*/

#define ADD_I        0x00000020		/* 00_100000   add		*/
#define ADDI_I       0x000000c8		/* 11_001000   addi		*/
#define ADDU_I       0x00000021		/* 00_100001   addu		*/
#define ADDUI_I      0x000000c9		/* 11_001001   addui		*/
#define AND_I        0x00000024		/* 00_100100   and		*/
#define ANDI_I       0x000000cc		/* 11_001100   andi		*/
#define BEQ_I        0x000000c4		/* 11_000100   beq		*/
#define BGEZ_I       0x00000041		/* 010_00001   bgez		*/
#define BGEZAL_I     0x00000051		/* 010_10001   bgezal		*/
#define BGTZ_I       0x000000c7		/* 11_000111   bgtz		*/
#define BLEZ_I       0x000000c6		/* 11_000110   blez		*/
#define BLTZ_I       0x00000040		/* 010_00000   bltz		*/
#define BLTZAL_I     0x00000050		/* 010_10000   bltzal		*/
#define BNE_I        0x000000c5		/* 11_000101   bne		*/
#define BREAK_I      0x0000000d		/* 00_001101   break		*/
#define J_I          0x000000c2		/* 11_000010   j		*/
#define JAL_I        0x000000c3		/* 11_000011   jal		*/
#define JALR_I       0x00000009		/* 00_001001   jalr		*/
#define JR_I         0x00000008		/* 00_001000   jr		*/
#define LB_I         0x000000e0		/* 11_100000   lb		*/
#define LBU_I        0x000000e4		/* 11_100100   lbu		*/
#define LH_I         0x000000e1		/* 11_100001   lh		*/
#define LHU_I        0x000000e5		/* 11_100101   lhu		*/
#define LUI_I        0x000000cf		/* 11_001111   lui		*/
#define LW_I         0x000000e3		/* 11_100011   lw		*/
#define MFC0_I       0x00000080		/* 10_000000   mfc0		*/
#define MTC0_I       0x00000084		/* 10_000100   mtc0		*/
#define NOR_I        0x00000027		/* 00_100111   nor		*/
#define OR_I         0x00000025		/* 00_100101   or		*/
#define ORI_I        0x000000cd		/* 11_001101   ori		*/
#define RFE_I        0x00000090		/* 10_010000   rfe		*/
#define SB_I         0x000000e8		/* 11_101000   sb		*/
#define SH_I         0x000000e9		/* 11_101001   sh		*/
#define SLEEP_I      0x0000000f		/* 00_001111   sleep		*/
#define SLL_I        0x00000000		/* 00_000000   sll		*/
#define SLLV_I       0x00000004		/* 00_000100   sllv		*/
#define SLT_I        0x0000002a		/* 00_101010   slt		*/
#define SLTI_I       0x000000ca		/* 11_001010   slti		*/
#define SLTU_I       0x0000002b		/* 00_101011   sltu		*/
#define SLTUI_I      0x000000cb		/* 11_001011   sltui		*/
#define SRL_I        0x00000002		/* 00_000010   srl		*/
#define SRLV_I       0x00000006		/* 00_000110   srlv		*/
#define SRA_I        0x00000003		/* 00_000011   sra		*/
#define SRAV_I       0x00000007		/* 00_000111   srav		*/
#define SUB_I        0x00000022		/* 00_100010   sub		*/
#define SUBU_I       0x00000023		/* 00_100011   subu		*/
#define SW_I         0x000000eb		/* 11_101011   sw		*/
#define SWAP_I       0x000000e7		/* 11_100111   swap		*/
#define SYSCALL_I    0x0000000c		/* 00_001100   syscall		*/
#define XOR_I        0x00000026		/* 00_100110   xor		*/
#define XORI_I       0x000000ce		/* 11_001110   xori		*/
#define MFHI_I       0x00000010		/* 00_010000   mfhi		*/
#define MTHI_I       0x00000011		/* 00_010001   mthi		*/
#define MFLO_I       0x00000012		/* 00_010010   mflo		*/
#define MTLO_I       0x00000013		/* 00_010011   mtlo		*/

	/* ###------------------------------------------------------### */
	/*   get the memory space reserved for the current instance.	*/
	/* Static memory is needed for:					*/
	/*     - registers and memory elements				*/
	/*     - last clock value to detect clock events		*/
	/* ###------------------------------------------------------### */

  vhx_getvalu (pt_lkdins, &pt_mem, sizeof (struct local_var), init_r3000);

	/* ###------------------------------------------------------### */
	/*    read the external clock 					*/
	/* ###------------------------------------------------------### */

  ck = vhx_readinput (pt_lkdins, PI_ck , 1);

	/* ###------------------------------------------------------### */
	/*    check for the falling edge of the clock			*/
	/* ###------------------------------------------------------### */

  if ((ck == 0) && (pt_mem->LAST_CK == 1))
    {
    pt_mem->LAST_CK = 0;

	/* ###------------------------------------------------------### */
	/*    read hardware interrupt request lines and set the		*/
	/* interrupt request bits of the cause register			*/
	/* ###------------------------------------------------------### */

    it_n          = vhx_readinput (pt_lkdins, PI_it_n, 6);
    pt_mem->CAUSE = (pt_mem->CAUSE & 0xffff03ff) | ((~it_n & 0x0000001f) << 10);

	/* ###------------------------------------------------------### */
	/*    in case of reset initialize the next instruction address	*/
	/* and the status register					*/
	/* ###------------------------------------------------------### */

    vhx_writeout (i_ftch, pt_lkdins, PO_i_ftch, 1);

    if (pt_mem->IFC_BUSY == 0)
      vhx_writeout (pt_mem->NEXTPC, pt_lkdins, PO_i_a, 32);

    reset_n = vhx_readinput (pt_lkdins, PI_reset_n, 1);

    if (reset_n == 0)
      {
      pt_mem->IFC_BUSY   = 1         ;
      pt_mem->NEXTPC     = RESET_A   ;
      pt_mem->NEXTSR     = 0x00400000;
      pt_mem->MEM_OPCOD  = ADDUI_I   ;
      pt_mem->MEM_RD_NBR = 0         ;
      pt_mem->MEM_OP     = OP_R2R    ;
      pt_mem->INST       = NOP_I     ;
      pt_mem->OPCOD      = ADDUI_I   ;
      pt_mem->RS_NBR     = 0         ;
      pt_mem->RT_NBR     = 0         ;
      pt_mem->RD_NBR     = 0         ;
      pt_mem->SHORT_I    = 0         ;
      pt_mem->SHIFT_AM   = 0         ;
      pt_mem->BRNCH_DEL  = 0         ;

      vhx_writeout (0         , pt_lkdins, PO_d_atype,  2);
      vhx_writeout (0         , pt_lkdins, PO_d_lock ,  1);
      vhx_writeout (D_NOACCESS, pt_lkdins, PO_d_rq   ,  1);
      }
    else
      {
	/* ###------------------------------------------------------### */
	/*    read the input ports and set the default values of	*/
	/* internal and output variables.				*/
	/* ###------------------------------------------------------### */

      if (pt_mem->MEM_OP == OP_LOAD)
        {
        d_in  = vhx_readinput (pt_lkdins, PI_d_in , 32);
        d_frz = vhx_readinput (pt_lkdins, PI_d_frz,  1);

	/* ###------------------------------------------------------### */
	/*    memory access and register file write back		*/
	/*								*/
	/*      - if the previous instruction was a load operation,	*/
	/*        read the data, align and make sign extension		*/
	/*      - complete the execution of instructions writing the	*/
	/*        result in the apropriate register			*/
	/* ###------------------------------------------------------### */

        if (d_frz == 1)
          pt_mem->MEM_BUSY = 1;
        else
          {
/*--------
printf ("\t\t-------- d_in  : %.8x ----------\n\n", d_in );
---------*/
          pt_mem->MEM_BUSY = 0               ;
          pt_mem->RD_BNBR2 = pt_mem->RD_BNBR1;
          pt_mem->RD_BNBR1 = (0x00000001 << pt_mem->MEM_RD_NBR) & 0xfffffffe;

          switch (pt_mem->MEM_OPCOD)
            {
	/* ###------------------------------------------------------### */
	/*    for a load word just read the data			*/
	/* ###------------------------------------------------------### */

            case LW_I:
              result = d_in;
              break;

	/* ###------------------------------------------------------### */
	/*    for a load half word unsigned, read the data, then align	*/
	/* and mask the sign.						*/
	/* ###------------------------------------------------------### */

            case LHU_I:
              if ((pt_mem->RESULT & 0x00000002) != 0)
                result = (d_in >> 16) & 0x0000ffff;
              else
                result = (d_in      ) & 0x0000ffff;
              break;

	/* ###------------------------------------------------------### */
	/*    for a load half word signed, read the data, then align	*/
	/* and extend the sign.						*/
	/* ###------------------------------------------------------### */

            case LH_I:
              if ((pt_mem->RESULT & 0x00000002) != 0)
                result = (d_in >> 16) & 0x0000ffff;
              else
                result = (d_in      ) & 0x0000ffff;

              if ((result & 0x00008000) != 0)
                result |= 0xffff0000;
              break;

	/* ###------------------------------------------------------### */
	/*    for a load byte unsigned, read the data, then align	*/
	/* and mask the sign.						*/
	/* ###------------------------------------------------------### */

            case LBU_I:
              switch (pt_mem->RESULT & 0x00000003)
                {
                case 3: result = (d_in >> 24) & 0x000000ff; break;
                case 2: result = (d_in >> 16) & 0x000000ff; break;
                case 1: result = (d_in >>  8) & 0x000000ff; break;
                case 0: result = (d_in      ) & 0x000000ff; break;
                }
              break;

	/* ###------------------------------------------------------### */
	/*    for a load byte signed, read the data, then align and	*/
	/* extend the sign.						*/
	/* ###------------------------------------------------------### */

            case LB_I:
              switch (pt_mem->RESULT & 0x00000003)
                {
                case 3: result = (d_in >> 24) & 0x000000ff; break;
                case 2: result = (d_in >> 16) & 0x000000ff; break;
                case 1: result = (d_in >>  8) & 0x000000ff; break;
                case 0: result = (d_in      ) & 0x000000ff; break;
                }
              if ((result & 0x00000080) != 0)
                result |= 0xffffff00;
              break;
            }
          }
        }
      else
        {
	/* ###------------------------------------------------------### */
	/*    for all other instructions just transfer the result	*/
	/* ###------------------------------------------------------### */

        pt_mem->RD_BNBR2 =   pt_mem->RD_BNBR1 |
                           ((0x00000001 << pt_mem->MEM_RD_NBR) & 0xfffffffe);
        pt_mem->RD_BNBR1 = 0             ;
        result           = pt_mem->RESULT;
        }

	/* ###------------------------------------------------------### */
	/*    save local variables					*/
	/*      - write back into the register file			*/
	/* ###------------------------------------------------------### */

      if (pt_mem->MEM_RD_NBR != 0)
        pt_mem->R [pt_mem->MEM_RD_NBR] = result;






      if (pt_mem->MEM_BUSY == 0)
        {
	/* ###------------------------------------------------------### */
	/*    instruction fetch, decode and execute			*/
	/* ###------------------------------------------------------### */

	/* ###------------------------------------------------------### */
	/*   read a new instruction if the last instruction has been	*/
	/* executed (the Instruction Fetch stage is not busy) and if	*/
	/* the memory stage is ready to receive a new instruction	*/
	/* ###------------------------------------------------------### */

        if (pt_mem->IFC_BUSY == 0)
          {
          new_inst = vhx_readinput (pt_lkdins, PI_i, 32);
/*---------
printf ("\t\t-------- new_inst : %.8x ----------\n\n", new_inst);
---------*/

	/* ###------------------------------------------------------### */
	/*   extract from the instruction :				*/
	/*     - the source registers number				*/
	/*     - the shift amount					*/
	/*     - the          long  immediate				*/
	/*     - the unsigned short immediate				*/
	/*     - the          short immediate				*/
	/* ###------------------------------------------------------### */

          rs_nbr   = (new_inst & 0x03e00000) >> 21;
          rt_nbr   = (new_inst & 0x001f0000) >> 16;
          shift_am = (new_inst & 0x000007c0) >>  6;
        
          if ((new_inst & 0x00008000) != 0)
            short_i = 0xffff0000 | (new_inst & 0x0000ffff);
          else
            short_i =              (new_inst & 0x0000ffff);

	/* ###------------------------------------------------------### */
	/*   extract from the instruction :				*/
	/*     - the operation code					*/
	/* ###------------------------------------------------------### */

          switch (new_inst & 0xfc000000)
            {
            case 0x00000000:				/* SPECIAL	*/
              opcod =                new_inst & 0x0000003f        ;
              break;
            case 0x04000000:				/* BCOND	*/
              opcod = 0x00000040 | ((new_inst & 0x001f0000) >> 16);
              break;
            case 0x40000000:				/* COP0		*/
              opcod = 0x00000080 | ((new_inst & 0x03e00000) >> 21);
              if ((opcod & 0x00000010) != 0)
                opcod &= 0x000000f0;
              break;
            default:					/* others	*/
              opcod = 0x000000c0 | ((new_inst & 0xfc000000) >> 26);
              break;
            }

	/* ###------------------------------------------------------### */
	/*   from the instruction extract the destination register	*/
	/*								*/
	/*   the destination number is located at 15-11 for:		*/
	/*								*/
	/*      add    : 00_100000, addu : 00_100001, and : 00_100100,	*/
	/*      jalr   : 00_001001, jr   : 00_001000, nor : 00_100111,	*/
	/*      or     : 00_100101, sleep: 00_001111, sll : 00_000000,	*/
	/*      sllv   : 00_000100, slt  : 00_101010, sltu: 00_101011,	*/
	/*      srl    : 00_000010, srlv : 00_000110, sra : 00_000011,	*/
	/*      srav   : 00_000111, sub  : 00_100010, subu: 00_100011,	*/
	/*      syscall: 00_001100, xor  : 00_100110, mfhi: 00_010000,	*/
	/*      mflo   : 00_010010					*/
	/*								*/
	/*   This cover all instructions which's OPCOD begin with 00	*/
	/*   except:							*/
	/*      break  : 00_001101, mthi : 00_010001, mtlo: 00_010011	*/
	/* ###------------------------------------------------------### */

          rd_nbr = 0;

          if (((opcod & 0x000000c0) == 0     ) && (opcod != BREAK_I) &&
              ( opcod               != MTHI_I) && (opcod != MTLO_I )   )
            rd_nbr = (new_inst & 0x0000f800) >> 11;

	/* ###------------------------------------------------------### */
	/*   from the instruction extract the destination register	*/
	/*								*/
	/*   the destination number is located at 20-16 for:		*/
	/*								*/
	/*      addi : 11_001000, addui: 11_001001, andi : 11_001100,	*/
	/*      lb   : 11_100000, lbu  : 11_100100, lh   : 11_100001,	*/
	/*      lhu  : 11_100101, lui  : 11_001111, lw   : 11_100011,	*/
	/*      ori  : 11_001101, slti : 11_001010, sltui: 11_001011,	*/
	/*      swap : 11_100111, xori : 11_001110, mfc0 : 10_000000	*/
	/*								*/
	/*   This cover all instructions which's OPCOD begin with 11	*/
	/*   except:							*/
	/*      beq  : 11_000100, bgtz : 11_000111, blez : 11_000110,	*/
	/*      bne  : 11_000101, j    : 11_000010, jal  : 11_000011,	*/
	/*      sb   : 11_101000, sh   : 11_101001, sw   : 11_101011	*/
	/*   and include MFC0						*/
	/* ###------------------------------------------------------### */

          if ((((opcod & 0x000000c0) == 0x000000c0) && 
               ((opcod & 0x00000038) != 0x00000000) &&
               ((opcod & 0x000000e8) != 0x000000e8)   ) ||
              (  opcod               == MFC0_I        )   )
            rd_nbr = (new_inst & 0x001f0000) >> 16;

	/* ###------------------------------------------------------### */
	/*   from the instruction extract the destination register	*/
	/*								*/
	/*   the destination number is 31 for:				*/
	/*      bgezal: 010_10001, bltzal: 010_10000, jal  : 11_000011,	*/
	/* ###------------------------------------------------------### */

          if ((opcod == BGEZAL_I) || (opcod == BLTZAL_I) || (opcod == JAL_I))
            rd_nbr = 31;
          }
        else
          {
          if (pt_mem->IFC_BUSY == 1)
            {
            new_inst = pt_mem->INST    ;
            opcod    = pt_mem->OPCOD   ;
            rs_nbr   = pt_mem->RS_NBR  ;
            rt_nbr   = pt_mem->RT_NBR  ;
            rd_nbr   = pt_mem->RD_NBR  ;
            short_i  = pt_mem->SHORT_I ;
            shift_am = pt_mem->SHIFT_AM;
            }
          }

	/* ###------------------------------------------------------### */
	/*    execute the current instruction:				*/
	/*      - read variables					*/
	/*      - set the default values of internal and output		*/
	/*        variables						*/
	/*      - by default, increment the program counter		*/
	/* ###------------------------------------------------------### */

        if (pt_mem->IFC_BUSY <= 1)
          {
          rs_bnbr        = 0x00000001 << rs_nbr  ;
          rt_bnbr        = 0x00000001 << rt_nbr  ;
          rd_bnbr1       = pt_mem->RD_BNBR1      ;
          rd_bnbr2       = pt_mem->RD_BNBR2      ;
          nextpc         = pt_mem->NEXTPC     + 4;
          s_oper         = pt_mem->R [rs_nbr]    ;
          t_oper         = pt_mem->R [rt_nbr]    ;
          pt_mem->MEM_OP = OP_R2R;

	/* ###------------------------------------------------------### */
	/*    execute the current instruction:				*/
	/*      - check if source operands are availables and set the	*/
	/*        busy bit						*/
	/*      - compute the result					*/
	/* ###------------------------------------------------------### */

          switch (opcod)
            {
            case ADD_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper + t_oper;
                if (((( (s_oper & t_oper) & ~result) |
                      (~(s_oper | t_oper) &  result)  ) >> 31) != 0)
                  exc_rq = OVR_EXC;
                }
              break;

            case ADDI_I    :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper + short_i;
                if (((( (s_oper & short_i) & ~result) |
                      (~(s_oper | short_i) &  result)  ) >> 31) != 0)
                  exc_rq = OVR_EXC;
                }
              break;

            case ADDU_I    :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper + t_oper;
                }
              break;

            case ADDUI_I   :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper + short_i;
                }
              break;

            case AND_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper & t_oper;
                }
              break;

            case ANDI_I    :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper & (new_inst & 0x0000ffff);
                }
              break;

            case BEQ_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                if (s_oper == t_oper)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BGEZ_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1         ;
                if (s_oper >= 0)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BGEZAL_I  :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1         ;
                result           = nextpc    ;
                if (s_oper >= 0)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BGTZ_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                if (s_oper > 0)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BLEZ_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                if (s_oper <= 0)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BLTZ_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                if (s_oper < 0)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BLTZAL_I  :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                result           = nextpc;
                if (s_oper < 0)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BNE_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                if (s_oper != t_oper)
                  nextpc += (short_i << 2) - 4;
                }
              break;

            case BREAK_I   :
              exc_rq = BREAK_EXC;
              break;

            case J_I       :
              pt_mem->IFC_BUSY = 0         ;
              pt_mem->MEM_OP   = OP_BRANCH ;
              pt_mem->REDOPC   = nextpc - 8;
              b_delay          = 1;
              nextpc           = ( (nextpc - 4) & 0xf0000000       ) |
                                 (( new_inst    & 0x03ffffff) <<  2);
              break;

            case JAL_I     :
              pt_mem->IFC_BUSY = 0         ;
              pt_mem->MEM_OP   = OP_BRANCH ;
              pt_mem->REDOPC   = nextpc - 8;
              b_delay          = 1;
              result           = nextpc;
              nextpc           = ( (nextpc - 4) & 0xf0000000       ) |
                                 (( new_inst    & 0x03ffffff) <<  2);
              break;

            case JALR_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                result           = nextpc;
                nextpc           = s_oper;
                }
              break;

            case JR_I      :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_BRANCH ;
                pt_mem->REDOPC   = nextpc - 8;
                b_delay          = 1;
                nextpc           = s_oper;
                }
              break;

            case LB_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_LOAD;
                result           = s_oper + short_i;
                d_rq             = D_ACCESS        ;
                d_atype          = D_READ          ;
                if (((result >> 31) & (pt_mem->NEXTSR >> 1)) != 0)
                  {
                  exc_rq          = ADEL_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case LBU_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_LOAD;
                result           = s_oper + short_i;
                d_rq             = D_ACCESS        ;
                d_atype          = D_READ          ;
                if (((result >> 31) & (pt_mem->NEXTSR >> 1)) != 0)
                  {
                  exc_rq          = ADEL_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case LH_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_LOAD;
                result           = s_oper + short_i;
                d_rq             = D_ACCESS        ;
                d_atype          = D_READ          ;
                if ((((result >> 31        ) & (pt_mem->NEXTSR >> 1)) != 0) ||
                    (( result &  0x00000001                         ) != 0)   )
                  {
                  exc_rq          = ADEL_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case LHU_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_LOAD;
                result           = s_oper + short_i;
                d_rq             = D_ACCESS        ;
                d_atype          = D_READ          ;
                if ((((result >> 31        ) & (pt_mem->NEXTSR >> 1)) != 0) ||
                    (( result &  0x00000001                         ) != 0)   )
                  {
                  exc_rq          = ADEL_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case LUI_I     :
              pt_mem->IFC_BUSY = 0;
              result           = (short_i << 16);
              break;

            case LW_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_LOAD;
                result           = s_oper + short_i;
                d_rq             = D_ACCESS        ;
                d_atype          = D_READ          ;
                if ((((result >> 31        ) & (pt_mem->NEXTSR >> 1)) != 0) ||
                    (( result &  0x00000003                         ) != 0)   )
                  {
                  exc_rq          = ADEL_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case MFC0_I    :
              if ((pt_mem->NEXTSR & 0x10000002) == 0x00000002)
                exc_rq = CPU_EXC;
              else
                {
                pt_mem->IFC_BUSY = 0;
                switch ((new_inst & 0x0000f800) >> 11)
                  {
                  case CAUSE_S  : result = pt_mem->CAUSE  ; break;
                  case BADVADR_S: result = pt_mem->BADVADR; break;
                  case EPC_S    : result = pt_mem->EPC    ; break;
                  case STATUS_S : result = pt_mem->NEXTSR ; break;
                  }
                }
              break;

            case MTC0_I    :
              if (((          rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                if ((pt_mem->NEXTSR & 0x10000002) == 0x00000002)
                  exc_rq = CPU_EXC;
                else
                  {
                  pt_mem->IFC_BUSY = 0;
                  switch ((new_inst & 0x0000f800) >> 11)
                    {
                    case CAUSE_S :
                      pt_mem->CAUSE = (pt_mem->CAUSE & 0xfffffcff) |
                                      (t_oper        & 0x00000300);
                      break;
                    case STATUS_S:
                      pt_mem->NEXTSR = t_oper;
                      break;
                    }
                  }
                }
              break;

            case NOR_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result = ~ (s_oper | t_oper);
                }
              break;

            case OR_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result = (s_oper | t_oper);
                }
              break;

            case ORI_I     :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result = s_oper | (new_inst & 0x0000ffff);
                }
              break;

            case RFE_I     :
              if ((pt_mem->NEXTSR & 0x10000002) == 0x00000002)
                exc_rq = CPU_EXC;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->NEXTSR = ( pt_mem->NEXTSR & 0xfffffff0      ) |
                                 ((pt_mem->NEXTSR & 0x0000003c) >> 2);
                }
              break;

            case SB_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_STORE        ;
                result           = s_oper + short_i;
                d_out            = t_oper          ;
                d_rq             = D_ACCESS        ;
                d_atype          = D_WRITEB        ;
                if (((result >> 31) & (pt_mem->NEXTSR >> 1)) != 0)
                  {
                  exc_rq          = ADES_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case SH_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_STORE        ;
                result           = s_oper + short_i;
                d_out            = t_oper          ;
                d_rq             = D_ACCESS        ;
                d_atype          = D_WRITEH        ;
                if ((((result >> 31        ) & (pt_mem->NEXTSR >> 1)) != 0) ||
                    (( result &  0x00000001                         ) != 0)   )
                  {
                  exc_rq          = ADES_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case SLEEP_I   :
              break;

            case SLL_I     :
              if (((          rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result = (t_oper << shift_am);
                }
              break;

            case SLLV_I    :
              if (((          rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result = (t_oper << (s_oper & 0x0000001f));
                }
              break;

            case SLT_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                if (s_oper < t_oper)
                  result = 1;
                else
                  result = 0;
                }
              break;

            case SLTI_I    :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                if (s_oper < short_i)
                  result = 1;
                else
                  result = 0;
                }
              break;

            case SLTU_I    :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                if (((unsigned int) s_oper) < ((unsigned int) t_oper))
                  result = 1;
                else
                  result = 0;
                }
              break;

            case SLTUI_I   :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                if (((unsigned int) s_oper) < ((unsigned int) short_i))
                  result = 1;
                else
                  result = 0;
                }
              break;

            case SRL_I     :
              if (((          rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = (t_oper     >>  shift_am     ) &
                                   (0x7fffffff >> (shift_am - 1)) ;
                }
              break;

            case SRLV_I    :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result = (t_oper     >>  (s_oper & 0x0000001f)     )  &
                         (0x7fffffff >> ((s_oper & 0x0000001f) - 1));
                }
              break;

            case SRA_I     :
              if (((          rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                if ((t_oper & 0x80000000) != 0)
                  result = (  t_oper     >>  shift_am      )  |
                           (~(0x7fffffff >> (shift_am - 1)));
                else
                  result = (  t_oper     >>  shift_am      )  &
                           (  0x7fffffff >> (shift_am - 1) );
                }
              break;

            case SRAV_I    :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                if ((t_oper & 0x80000000) != 0)
                  result = (  t_oper     >>  (s_oper & 0x0000001f)      )  |
                           (~(0x7fffffff >> ((s_oper & 0x0000001f) - 1)));
                else
                  result = (  t_oper     >>  (s_oper & 0x0000001f)      )  &
                           (  0x7fffffff >> ((s_oper & 0x0000001f) - 1) );
                }
              break;

            case SUB_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper - t_oper;
                if (((( (t_oper & result) & ~s_oper) |
                      (~(t_oper | result) &  s_oper)  ) >> 31) != 0)
                  exc_rq = OVR_EXC;
                }
              break;

            case SUBU_I    :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper - t_oper;
                }
              break;

            case SW_I      :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_STORE        ;
                result           = s_oper + short_i;
                d_out            = t_oper          ;
                d_rq             = D_ACCESS        ;
                d_atype          = D_WRITEW        ;
                if ((((result >> 31        ) & (pt_mem->NEXTSR >> 1)) != 0) ||
                    (( result &  0x00000003                         ) != 0)   )
                  {
                  exc_rq          = ADES_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case SWAP_I    :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->MEM_OP   = OP_SWAP         ;
                result           = s_oper + short_i;
                d_out            = t_oper          ;
                d_rq             = D_ACCESS        ;
                d_atype          = D_SWAPRQ        ;
                d_lock           = 1               ;
                if ((((result >> 31        ) & (pt_mem->NEXTSR >> 1)) != 0) ||
                    (( result &  0x00000003                         ) != 0)   )
                  {
                  exc_rq          = ADES_EXC;
                  pt_mem->BADVADR = result  ;
                  }
                }
              break;

            case SYSCALL_I :
              exc_rq = SYSCALL_EXC;
              break;

            case XOR_I     :
              if (((rs_bnbr | rt_bnbr) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper ^ t_oper;
                }
              break;

            case XORI_I    :
              if (((rs_bnbr          ) & (rd_bnbr1           )) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                result           = s_oper ^ (new_inst & 0x0000ffff);
                }
              break;

            case MFHI_I    :
              pt_mem->IFC_BUSY = 0;
              result           = pt_mem->HI;
              break;

            case MTHI_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->HI       = s_oper;
                }
              break;

            case MFLO_I    :
              pt_mem->IFC_BUSY = 0;
              result           = pt_mem->LO;
              break;

            case MTLO_I    :
              if (((rs_bnbr          ) & (rd_bnbr1 | rd_bnbr2)) != 0)
                pt_mem->IFC_BUSY = 1;
              else
                {
                pt_mem->IFC_BUSY = 0;
                pt_mem->LO       = s_oper;
                }
              break;

            default        :
              exc_rq = RESERVE_EXC;
              break;
            }

          if (exc_rq != 0)
            {
            vhx_writeout (D_NOACCESS, pt_lkdins, PO_d_rq, 1);

            pt_mem->MEM_OPCOD  = ADDUI_I;
            pt_mem->MEM_RD_NBR = 0      ;
            pt_mem->MEM_OP     = OP_R2R ;
            pt_mem->RESULT     = 0      ;
            pt_mem->IFC_BUSY   = 4      ;

            pt_mem->INST       = NOP_I  ;
            pt_mem->OPCOD      = ADDUI_I;
            pt_mem->RS_NBR     = 0      ;
            pt_mem->RT_NBR     = 0      ;
            pt_mem->RD_NBR     = 0      ;
            pt_mem->SHORT_I    = 0      ;
            pt_mem->SHIFT_AM   = 0      ;

            pt_mem->CAUSE  = ( pt_mem->CAUSE  & 0xffffffc3      ) |
                             ( exc_rq         & 0x0000003c      );
            pt_mem->NEXTSR = ( pt_mem->NEXTSR & 0xffffffc0      ) |
                             ((pt_mem->NEXTSR & 0x0000000f) << 2);

            if (pt_mem->BRNCH_DEL == 0)
              {
              pt_mem->EPC    = pt_mem->NEXTPC - 4;
              pt_mem->CAUSE &= 0x7fffffff;
              }
            else
              {
              pt_mem->EPC    = pt_mem->REDOPC;
              pt_mem->CAUSE |= 0x80000000;
              }

            if ((pt_mem->NEXTSR & 0x00400000) != 0)
              pt_mem->NEXTPC = BOOTHND_A;
            else
              pt_mem->NEXTPC = EXCPHND_A;
            }
          }

        if (exc_rq == 0)
          {
          if (((pt_mem->NEXTSR & 0x0000ff00 & pt_mem->CAUSE) != 0) &&
              ((pt_mem->NEXTSR & 0x00000001                ) != 0)   )
            {
            if ((pt_mem->IFC_BUSY == 0) && (pt_mem->MEM_OP != OP_BRANCH))
              {
              pt_mem->MEM_OPCOD  = opcod  ;
              pt_mem->MEM_RD_NBR = rd_nbr ;
              pt_mem->RESULT     = result ;
              pt_mem->NEXTPC     = nextpc ;
              pt_mem->BRNCH_DEL  = b_delay;

              if ((pt_mem->MEM_OP == OP_LOAD ) ||
                  (pt_mem->MEM_OP == OP_STORE) ||
                  (pt_mem->MEM_OP == OP_SWAP )   )
                {
                vhx_writeout (result , pt_lkdins, PO_d_a    , 32);
                vhx_writeout (d_atype, pt_lkdins, PO_d_atype,  2);

                if (pt_mem->MEM_OP != OP_LOAD)
                  vhx_writeout (d_out, pt_lkdins, PO_d_out, 32);
                }
              vhx_writeout (d_lock, pt_lkdins, PO_d_lock,  1);
              vhx_writeout (d_rq  , pt_lkdins, PO_d_rq  ,  1);
              }
            else
              {
              pt_mem->MEM_OPCOD  = ADDUI_I;
              pt_mem->MEM_RD_NBR = 0      ;
              pt_mem->MEM_OP     = OP_R2R ;
              pt_mem->RESULT     = 0      ;
              vhx_writeout (D_NOACCESS, pt_lkdins, PO_d_rq  ,  1);
              }

            pt_mem->IFC_BUSY   = 4      ;

            pt_mem->INST       = NOP_I  ;
            pt_mem->OPCOD      = ADDUI_I;
            pt_mem->RS_NBR     = 0      ;
            pt_mem->RT_NBR     = 0      ;
            pt_mem->RD_NBR     = 0      ;
            pt_mem->SHORT_I    = 0      ;
            pt_mem->SHIFT_AM   = 0      ;

            pt_mem->CAUSE  = ( pt_mem->CAUSE  & 0xffffffc3      );
            pt_mem->NEXTSR = ( pt_mem->NEXTSR & 0xffffffc0      ) |
                             ((pt_mem->NEXTSR & 0x0000000f) << 2);

            if (pt_mem->BRNCH_DEL == 0)
              {
              pt_mem->EPC    = pt_mem->NEXTPC - 4;
              pt_mem->CAUSE &= 0x7fffffff;
              }
            else
              {
              pt_mem->EPC    = pt_mem->REDOPC;
              pt_mem->CAUSE |= 0x80000000;
              }

            if ((pt_mem->NEXTSR & 0x00400000) != 0)
              pt_mem->NEXTPC = BOOTHND_A;
            else
              pt_mem->NEXTPC = EXCPHND_A;
            }
          else
            {
            if (pt_mem->IFC_BUSY == 0)
              {
              pt_mem->MEM_OPCOD  = opcod  ;
              pt_mem->MEM_RD_NBR = rd_nbr ;
              pt_mem->RESULT     = result ;
              pt_mem->NEXTPC     = nextpc ;
              pt_mem->BRNCH_DEL  = b_delay;

              if ((pt_mem->MEM_OP == OP_LOAD ) ||
                  (pt_mem->MEM_OP == OP_STORE) ||
                  (pt_mem->MEM_OP == OP_SWAP )   )
                {
                vhx_writeout (result , pt_lkdins, PO_d_a    , 32);
                vhx_writeout (d_atype, pt_lkdins, PO_d_atype,  2);

                if (pt_mem->MEM_OP != OP_LOAD)
                  vhx_writeout (d_out  , pt_lkdins, PO_d_out  , 32);
                }
              vhx_writeout (d_lock , pt_lkdins, PO_d_lock ,  1);
              vhx_writeout (d_rq   , pt_lkdins, PO_d_rq   ,  1);
              }
            else
              {
              if (pt_mem->IFC_BUSY == 1)
                {
                pt_mem->MEM_OPCOD  = ADDUI_I;
                pt_mem->MEM_RD_NBR = 0      ;
                pt_mem->MEM_OP     = OP_R2R ;
                pt_mem->RESULT     = 0      ;

                pt_mem->INST       = new_inst  ;
                pt_mem->OPCOD      = opcod     ;
                pt_mem->RS_NBR     = rs_nbr    ;
                pt_mem->RT_NBR     = rt_nbr    ;
                pt_mem->RD_NBR     = rd_nbr    ;
                pt_mem->SHORT_I    = short_i   ;
                pt_mem->SHIFT_AM   = shift_am  ;
                }
              }
            }
          }

        if (pt_mem->IFC_BUSY > 1)
          pt_mem->IFC_BUSY --;
        }
      }

/*--------
display_var (pt_mem);
----------*/
    }
	/* ###------------------------------------------------------### */
	/*    check for rising edge of the clock			*/
	/*    update the last clock value				*/
	/* ###------------------------------------------------------### */

  if ((ck == 1) && (pt_mem->LAST_CK == 0))
    pt_mem->LAST_CK = 1;

  }

/* ###--------------------------------------------------------------###	*/
/* function	: init_r3000						*/
/* description	: initialization function				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------###	*/

static void init_r3000 (pt_lkdins, pt_mem)

struct lkdins    *pt_lkdins;
struct local_var *pt_mem   ;

  {
  pt_mem->LAST_CK   = 0         ;
  pt_mem->MEM_BUSY  = 0         ;
  pt_mem->IFC_BUSY  = 0         ;
  pt_mem->CAUSE    &= 0x0000ff3c;
  pt_mem->R [0]     = 0         ;
  }

/* ###--------------------------------------------------------------###	*/
/* function	: display_var						*/
/* description	: display local variables				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------###	*/

static void display_var (pt_mem)

struct local_var *pt_mem;

  {
  printf ("nxtpc: %.8x, nxtsr: %.8x, cause: %.8x, epc  : %.8x\n",
          pt_mem->NEXTPC, pt_mem->NEXTSR, pt_mem->CAUSE, pt_mem->EPC);

  printf ("i_bsy: %8.x, m_bsy: %8.x, b_dly: %8.x, redo : %.8x\n",
          pt_mem->IFC_BUSY, pt_mem->MEM_BUSY, pt_mem->BRNCH_DEL, pt_mem->REDOPC);

  printf ("rsnbr: %.8x, rtnbr: %.8x, rdnbr: %.8x\n",
          pt_mem->RS_NBR, pt_mem->RT_NBR, pt_mem->RD_NBR);

/*-----
  printf ("r[19]: %.8x, r[31]: %.8x\n", pt_mem->R[19], pt_mem->R[31]);

  printf ("r[ 1]: %.8x, r[26]: %.8x, r[30]: %.8x\n",
          pt_mem->R[ 1], pt_mem->R[26], pt_mem->R[30]);

  printf ("r[ 1]: %.8x, r[ 2]: %.8x, r[ 3]: %.8x, r[ 4]: %.8x\n",
          pt_mem->R[ 1], pt_mem->R[ 2], pt_mem->R[ 3], pt_mem->R[ 4]);
  printf ("r[ 5]: %.8x, r[ 6]: %.8x, r[ 7]: %.8x, r[ 8]: %.8x\n",
          pt_mem->R[ 5], pt_mem->R[ 6], pt_mem->R[ 7], pt_mem->R[ 8]);
  printf ("r[ 9]: %.8x, r[10]: %.8x, r[11]: %.8x, r[12]: %.8x\n",
          pt_mem->R[ 9], pt_mem->R[10], pt_mem->R[11], pt_mem->R[12]);

  printf ("lo   : %.8x, hi   : %.8x, badad: %.8x, memop: %.8x, ",
           pt_mem->LO, pt_mem->HI, pt_mem->BADVADR, pt_mem->MEM_OPCOD);

--------*/
  }
