/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : BooG - read all cells
 * Date    : 2000
 * Author  : Francois Donnet
 */


#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include MUT_H
#include ABL_H
#include MLO_H
#include ABE_H
#include MLU_H
#include ABV_H
#include "bog_normalize_register.h"
#include "bog_normalize_ARITY.h"
#include "bog_normalize_DAG.h"
#include "bog_normalize_simplify.h"
#include "bog_normalize_power.h"
#include "bog_lib_utils.h"
#include "bog_lib_cell.h"
#include "bog_lib_negativ.h"
#include "bog_lib_format.h"
#include "bog_lib_permute.h"
#include "bog_lib_specifications.h"
#include "bog_lib_matching.h"
#include "bog_lib_reader.h"


/***************************************************************************/
/*     return a list of files in directory path with the extension ext     */
/* if error in path then return NULL                                       */
/***************************************************************************/
extern chain_list *getfiles_with_ext (char *path, char *ext)
{
  DIR *dir;
  struct dirent *entry;
  char *filename;
  char *completefilename;
  int filenamelength, extensionlength, pathlength;
  chain_list *head;

  if (!path || !ext) return NULL;

  dir = opendir (path);
  if (!dir) return NULL;

  extensionlength = strlen (ext);
  pathlength = strlen (path); 
  head=NULL;
  
  while ((entry = readdir (dir)) != NULL) {

      filename = entry->d_name;
      filenamelength = strlen (filename);
     
      if (filenamelength <= extensionlength) continue;
      /* is extension of filename accepted */
      if (strncmp (filename + filenamelength - extensionlength,
                  ext, extensionlength)) continue;

      completefilename = mbkalloc(filenamelength + 1);
      sprintf(completefilename, "%s", filename);
      head = addchain (head, completefilename);
  }

  return head;
}


/******************************************************************************/
/*   for each aspect of abl, build a cell                                     */
/* return 1 if included in library                                            */
/******************************************************************************/
static int distribCell(befig_list* befig)
{
   chain_list *permutations=NULL,*permutations_val=NULL,*permutations_cnd=NULL;
   chain_list *chain, *chain_cnd, *chain_val;
   ptype_list* ptype;
   cell_list *cells_lib, *cell, *pred;
   char *old_cell;

   /*register*/
   if (befig->BEREG) {
   
      /*in negativ logic, propagate NOT*/
      befig->BEREG->BIABL->CNDABL=inv_oper(befig->BEREG->BIABL->CNDABL,0);
      befig->BEREG->BIABL->VALABL=inv_oper(befig->BEREG->BIABL->VALABL,0);
      /*minmize number of NOT*/
      befig->BEREG->BIABL->CNDABL=build_negativ(befig->BEREG->BIABL->CNDABL);
      befig->BEREG->BIABL->VALABL=build_negativ(befig->BEREG->BIABL->VALABL);
      /*flip-flop or latch*/
      ptype=getptype(befig->BEREG->BIABL->USER,ABL_STABLE);
      if (ptype) cells_lib=getcell_flip_flop_lib();
      else cells_lib=getcell_latch_lib();
      
      /*check the use to insert this cell*/ 
      for (cell=cells_lib; cell; cell=cell->NEXT) {
         /*new more precise than old -> put new before*/
         if (pattern_matching(befig->BEREG->BIABL->CNDABL,cell->CNDABL)
          && pattern_matching(befig->BEREG->BIABL->VALABL,cell->VALABL)) {
            /*patterns are equal*/
            if (pattern_matching(cell->CNDABL,befig->BEREG->BIABL->CNDABL)
             && pattern_matching(cell->VALABL,befig->BEREG->BIABL->VALABL)) {
               if (cell->AREA<getgenericarea(befig)) return 0;
               if (cell->AREA==getgenericarea(befig)) break; /*insert new*/
               /*remove old cells because they are bigger*/
               fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",cell->NAME);
               old_cell=cell->NAME;
               pred=NULL;
               for (cell=cells_lib ; cell; cell=cell->NEXT) {
                  if (cell->NAME==old_cell && pred) pred->NEXT=cell->NEXT;
                  else if (cell->NAME==old_cell) cells_lib=cell->NEXT;
                  else pred=cell;
               }
               if (ptype) setcell_flip_flop_lib(cells_lib);
               else setcell_latch_lib(cells_lib);
            }
            /*insert new*/
            break;
         }   
      }      
      
      /*give the list of different permutations*/
      permutations_cnd=different_abl(befig,befig->BEREG->BIABL->CNDABL);
      permutations_val=different_abl(befig,befig->BEREG->BIABL->VALABL);
      
      /*put in cell*/
      for (chain_cnd=permutations_cnd; chain_cnd; chain_cnd=chain_cnd->NEXT) {
      for (chain_val=permutations_val; chain_val; chain_val=chain_val->NEXT) {
         /*flip-flop or latch*/
         if (ptype) addCell_flip_flop(befig, dupablexpr(chain_cnd->DATA),
                     dupablexpr(chain_val->DATA));
         else addCell_latch(befig, dupablexpr(chain_cnd->DATA),
                     dupablexpr(chain_val->DATA));
      }
      }

   }


   /*bus*/
   else if (befig->BEBUS) {
      /*in negativ logic, propagate NOT*/
      befig->BEBUS->BIABL->CNDABL=inv_oper(befig->BEBUS->BIABL->CNDABL,0);
      befig->BEBUS->BIABL->VALABL=inv_oper(befig->BEBUS->BIABL->VALABL,0);
      /*minmize number of NOT*/
      befig->BEBUS->BIABL->CNDABL=build_negativ(befig->BEBUS->BIABL->CNDABL);
      befig->BEBUS->BIABL->VALABL=build_negativ(befig->BEBUS->BIABL->VALABL);
      cells_lib=getcell_tristate_lib();
      
      /*check the use to insert this cell*/ 
      for (cell=cells_lib; cell; cell=cell->NEXT) {
         /*new more precise than old -> put new before*/
         if (pattern_matching(befig->BEBUS->BIABL->CNDABL,cell->CNDABL)
          && pattern_matching(befig->BEBUS->BIABL->VALABL,cell->VALABL)) {
            /*patterns are equal*/
            if (pattern_matching(cell->CNDABL,befig->BEBUS->BIABL->CNDABL)
             && pattern_matching(cell->VALABL,befig->BEBUS->BIABL->VALABL)) {
               if (cell->AREA<getgenericarea(befig)) return 0;
               if (cell->AREA==getgenericarea(befig)) break; /*insert new*/
               /*remove old cells because they are bigger*/
               fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",cell->NAME);
               old_cell=cell->NAME;
               pred=NULL;
               for (cell=cells_lib ; cell; cell=cell->NEXT) {
                  if (cell->NAME==old_cell && pred) pred->NEXT=cell->NEXT;
                  else if (cell->NAME==old_cell) cells_lib=cell->NEXT;
                  else pred=cell;
               }
               setcell_tristate_lib(cells_lib);
            }
            /*insert new*/
            break;
         }   
      }      
      
      /*give the list of different permutations*/
      permutations_cnd=different_abl(befig,befig->BEBUS->BIABL->CNDABL);
      permutations_val=different_abl(befig,befig->BEBUS->BIABL->VALABL);
      /*put in cell*/
      for (chain_cnd=permutations_cnd; chain_cnd; chain_cnd=chain_cnd->NEXT) {
      for (chain_val=permutations_val; chain_val; chain_val=chain_val->NEXT) {
         addCell_bus(befig, dupablexpr(chain_cnd->DATA), 
            dupablexpr(chain_val->DATA));
      }
      }

   }
   
   /*logic*/
   else if (befig->BEOUT) {/*pure logical*/
      /*in negativ logic, propagate NOT*/
      befig->BEOUT->ABL=inv_oper(befig->BEOUT->ABL,0);
      /*minmize number of NOT*/
      befig->BEOUT->ABL=build_negativ(befig->BEOUT->ABL);
      cells_lib=getcell_logic_lib();
      
      /*check the use to insert this cell*/ 
      for (cell=cells_lib; cell; cell=cell->NEXT) {
         /*new more precise than old -> put new before*/
         if (pattern_matching(befig->BEOUT->ABL,cell->VALABL)) {
            /*patterns are equal*/
            if (pattern_matching(cell->VALABL,befig->BEOUT->ABL)) {
               if (cell->AREA<getgenericarea(befig)) return 0;
               if (cell->AREA==getgenericarea(befig)) break; /*insert new*/
               /*remove old cells because they are bigger*/
               fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",cell->NAME);
               old_cell=cell->NAME;
               pred=NULL;
               for (cell=cells_lib ; cell; cell=cell->NEXT) {
                  if (cell->NAME==old_cell && pred) pred->NEXT=cell->NEXT;
                  else if (cell->NAME==old_cell) cells_lib=cell->NEXT;
                  else pred=cell;
               }
               setcell_logic_lib(cells_lib);
            }   
            /*insert new*/
            break;
         }   
      }      
      
      /*give the list of different permutations*/
      permutations=different_abl(befig,befig->BEOUT->ABL);
      for (chain=permutations; chain; chain=chain->NEXT) {
         /*several abl for a same behavior!!! ex: and(or ...) == nor(nor...)*/
         addCell_logic(befig,chain->DATA);
      }      
   }
   
   /*free mem*/
   for (chain=permutations_cnd; chain; chain=chain->NEXT) {
      freeablexpr(chain->DATA);
   }
   for (chain=permutations_val; chain; chain=chain->NEXT) {
      freeablexpr(chain->DATA);
   }
   
   if (permutations_cnd) freechain(permutations_cnd);
   if (permutations_val) freechain(permutations_val);
   if (permutations) freechain(permutations);
   
   return 1;
}


/*****************************************************************************/
/*                   build the cell library                                  */
/*****************************************************************************/
extern void library_reader(char* cell_directory)
{
   char *name;
   chain_list* vbe_list, *chain;
   befig_list* befig;
   char* memo_work;
   char** memo_cata;
   char* new_cata[2];

   if (!cell_directory) {
      fprintf(stderr,"library_reader: no Cell directory\n");
      exit(1);
   }
   
   /*seek vbe files which describe the lib*/
   vbe_list=getfiles_with_ext (cell_directory, ".vbe");  
   
   /*search only in our library not to interfer with others*/
   memo_work=WORK_LIB;
   memo_cata=CATA_LIB;
   WORK_LIB=cell_directory;  
   CATA_LIB=new_cata;
   new_cata[0]=cell_directory;
   new_cata[1]=NULL;

   if (!vbe_list /*list of name*/) {
      fprintf(stderr,
              "Mapping Error: no cell in directory '%s'\n", 
              cell_directory?cell_directory:".");
      exit(1);
   }
   
   for (chain=vbe_list; chain; chain=chain->NEXT) {

      /*put off extension*/
      name=chain->DATA;
      name[strlen(name)-strlen(".vbe")]='\0';
      
      /* mode: BVL_KEEPAUX , BVL_TRACE ... */
      /*do not keep aux*/
      befig=vhdlloadbefig(NULL,name,'P');
      if (befig->ERRFLG) {
         fprintf(stderr,
                 "BEH: unknown error %d in %s.vbe\n",befig->ERRFLG,name);
         fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",name);
         continue;
      }

      /*detect power and forbid Vdd and Vss in expression*/
      if (!detect_power(befig)) {
         fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",befig->NAME);
         continue;
      }    

      /*simplify abl expression   -needed by format_register()- */
      simplify_expr(befig);

      /*remove stable             -needed by DAG_control()-  */
      if (!format_register(befig)) {
         fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",befig->NAME);
         continue;
      }    

      /*control cycles in befig and erase unused bebux and beaux*/
      if (!DAG_control(befig)) {
         fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",befig->NAME);
         continue;
      }    

      /*if constraints are respected*/
      if (!format_cell(befig)) {
         fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",befig->NAME);
         continue;
      }    
   
      /*put the locon of lofig in the same order than bepor of befig*/
      if (sortbepor(befig)) {
         fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",befig->NAME);
         continue;
      }    
      
      /*put arity oper*/
      put_arity(befig);
      
      /*memorize cell if not already exists*/
      if (!distribCell(befig)) {
        fprintf(stderr,"Mapping Warning: Cell '%s' unused\n",befig->NAME);
      }
   
   }

   freechain(vbe_list);
   
   /*relax old env*/
   WORK_LIB=memo_work;
   CATA_LIB=memo_cata;
}

